/*
 * rssh.c - restricted shell for ssh to allow scp or sftp only
 * 
 * Copyright 2002 Derek D. Martin <code@pizzashack.org>.
 *
 * This program is licensed under a BSD-style license, as follows: 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *	notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *	notice, this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Credits
 *
 * Thanks to the following people for pointing out some lameness in early
 * versions of the program:
 *
 * Mike O'Donnell
 * Derek Atkins
 * Frank Louwers
 * Ivo De Decker
 *
 */

/* SYSTEM INCLUDES */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <libgen.h>
#include <wordexp.h>
#include <syslog.h>
#include <pwd.h>
#include <sys/types.h>

/* LOCAL INCLUDES */
#include "pathnames.h"
#include "log.h"
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

/* MACRO DEFINITIONS */
#define _ALLOW_SCP 1
#define _ALLOW_SFTP (1 << 1)

/* FILE SCOPE FUNCTION DECLARATIONS */
void fail( int flags, int argc, char **argv );
char **build_arg_vector( char *str );
char *get_username( void );

/* GLOBAL VARIABLES */
extern int errno;

/* FILE SCOPE VARIABLES */
static char *progname;
static char *username;
static char *version = 
	"rssh 1.0.3 (c) 2002 Derek D. Martin <code@pizzashack.org>";

/* MAIN PROGRAM */
int main( int argc, char **argv )
{

	char	**argvec;
	int	capabilities = 0;

	/* see how we were called, and set up logging */
	if ( !(username = get_username()) )
		/* this SHOULD never happen */
		username = "unknown user!";
	progname = strdup(log_make_ident(basename(argv[0])));
	log_set_priority(LOG_DAEMON,LOG_ERR);
	log_open();

	/* to chroot, or not to chroot... */
	/* coming soon.  :) */
	   
	/*  check progname for scp or sftp to see what's allowed */
	if ( !strcmp(progname, "scpsh") ) capabilities = _ALLOW_SCP;
	if ( !strcmp(progname, "sftpsh") ) capabilities = _ALLOW_SFTP;
	if ( !strcmp(progname, "rssh") ) 
		capabilities = _ALLOW_SFTP | _ALLOW_SCP;

	/* q&d arg count check */
	if ( argc == 2 && !( strcmp(argv[1], "-v")) ){
		printf("%s\n\n", version);
		exit(0);
	}

	if ( argc < 3 ) fail(capabilities, argc, argv);

	/* if first arg is anything but -c, it's no good */
	if ( strcmp("-c", argv[1]) ) fail(capabilities, argc, argv);

	/* convert argv[2] into an arg vector suitable for execvp() */
	argvec = build_arg_vector(argv[2]);

	/* check to see if we got an allowed command */
	if ( 
	     !( 
		( !strcmp(argvec[0], _PATH_SFTP_SERVER) && 
		  (capabilities & _ALLOW_SFTP) )  || 
		( !strcmp(argvec[0], "scp") && (capabilities & _ALLOW_SCP) )
	      )
	   )
		fail(capabilities, argc, argv);

	if ( !strcmp(argvec[0], "scp") )
		progname = _PATH_SCP;
	else
		progname = _PATH_SFTP_SERVER;

	/* if all that passed, exec the relevant command */
	execv(progname, argvec);

	/* we only get here if the exec fails */
	fprintf(stderr, "%s: excecvp() failed.  ", progname);
	log_msg("execv failed: cmd line %s", argv[2]);

	switch (errno){
	case EACCES:
	case ENOTDIR:
	case ENOENT:
		fprintf(stderr, "%s: %s is not an executable file, or "
			"permission denied.\n\n", basename(argv[0]), 
			argvec[0]);
		break;
	case EPERM:
		/* this shouldn't happen, as we don't run SUID */
		fprintf(stderr, "%s: FS mounted nosuid or process is being "
			"traced\n (and you are not root)\n\n", 
			basename(argv[0]));
		break;
	default:
			fprintf(stderr, "an unknown error occurred.\n\n");
	}
	
	exit(1);
}


char **build_arg_vector( char *str )
{

	wordexp_t   	result;
	int		retc;

	if ( (retc = wordexp(str, &result, WRDE_NOCMD)) ){
		switch( retc ){
		case WRDE_BADCHAR:
		case WRDE_CMDSUB:
			fprintf(stderr, "%s: bad characters in arguments\n", 
				progname);
			log_msg("user %s used bad chars in command",
				username);
			break;
		default:
			fprintf(stderr, "%s: error expanding arguments\n", 
				progname);
			log_msg("error expanding arguments for user %s",
				username);
		}

		exit(1);
	}

	return result.we_wordv;
}


void fail( int flags, int argc, char **argv )
{

	char *cmd;

	/* determine which commands are usable for error message */
	switch (flags){
	case (_ALLOW_SCP | _ALLOW_SFTP):
		cmd = "scp or sftp";
		break;
	case _ALLOW_SCP:
		cmd = "scp";
		break;
	case _ALLOW_SFTP:
		cmd = "sftp";
		break;
	default:
		fprintf(stderr, "\n\nrssh error - unknown capabilities.\n\n  "
			"rssh is probably installed incorrectly.  Please see "
			"your system administrator.\n\n");
		log_msg("bad capabilities - check installation of rssh");
		exit(1);
	}

	/* print error message to user and log attempt */
	fprintf(stderr, "\nThis account is restricted to %s only.  If you "
		"believe\n this is in error, please contact your system "
		"administrator.\n\n", cmd);
	if ( argc < 3 )
		log_msg("user %s attempted to log in with a shell",
			username);
	else{
		log_msg("user %s attempted to execute forbidden commands",
			username);
		log_msg("command: %s", argv[2]);
	}

	exit(0);
}

char *get_username( void )
{

	struct passwd *entry;
	char *retc;

	entry = getpwuid( getuid() );
	retc = strdup(entry->pw_name);
	return retc;
}
